functor LatexWr (structure Wr: WR) : LATEXWR =
struct

structure Wr = Wr

val verbatimcommand = "verbatim"

datatype Mode = 
        OUTER     	(* before first character written *)
      | TEXTONLY 	(* text only mode -- no latex formatting output; 
      			   mode changes are ignored *)
      | TEXT 		(* latex mode, formatting text *)
      | AFTERTEXT 	(* latex mode, about to format code *)
      | VERBATIM	(* latex mode, formatting code *)

val LATEX_MODE = ref false;

type T = 
      { w : Wr.Wr,
        m : Mode ref
	}
		 
fun enterverb {w=w,m=m} = 
  (Wr.write_wr w ("\\begin{" ^ verbatimcommand ^ "}\n");
   m := VERBATIM)

fun leaveverb {w=w,m=m} = 
  (Wr.write_wr w ("\\end{" ^ verbatimcommand ^ "}\n");
   m := TEXT)

exception LatexFlagsImproperlyNested

fun close_lw lw = let 
  val {w=w,m=m} = lw
  in
    case (!m) of
        OUTER => ()
      | TEXT => raise LatexFlagsImproperlyNested
      | TEXTONLY => ()
      | AFTERTEXT => ()
      | VERBATIM => leaveverb lw ;
    Wr.flush w;
    Wr.close w
  end

fun write_lw lw s = let
  val {w=w,m=m} = lw 
  in
    case (!m) of 
	OUTER => if !LATEX_MODE
		   then enterverb lw
		   else m := TEXTONLY
      | TEXT => ()
      | TEXTONLY => ()
      | AFTERTEXT => enterverb lw
      | VERBATIM => () ;
    Wr.write_wr w s
  end

fun begintext lw = let
  val {w=w,m=m} = lw 
  in
    case (!m) of 
	OUTER => m := TEXT
      | TEXT => raise LatexFlagsImproperlyNested
      | TEXTONLY => ()
      | AFTERTEXT => m := TEXT
      | VERBATIM => leaveverb lw
  end

fun endtext lw = let
  val {w=w,m=m} = lw 
  in
    case (!m) of 
	OUTER => raise LatexFlagsImproperlyNested
      | TEXT => m := AFTERTEXT
      | TEXTONLY => ()
      | AFTERTEXT => raise LatexFlagsImproperlyNested
      | VERBATIM => raise LatexFlagsImproperlyNested
  end

fun to_wr underlying_wr = let 
  val lw = { w = underlying_wr,
             m = ref(OUTER) }
  val wr = Wr.to_fn 
      	      (fn s => write_lw lw s)
      	      (fn _ => close_lw lw)
      	      (fn _ => Wr.flush underlying_wr)
  in (wr,lw)
  end

end 

