functor Typ(
	    structure Globals: GLOBALS
	    ) : TYP
	    = struct

structure Globals = Globals
open Globals
open Pp

datatype K = 
	    STAR
	  | KARROW of K * K

and T = 
	    TVAR of int
	  | ABST of Id.T * K * T
	  | APPT of T * T
	  | ARROW of T * T
	  | ALL of Id.T * T * T
	  | SOM of Id.T * T * T
	  | TOP of K 
	  | RCD of (Id.T * T) list


(* ---------------------------  Kinds  -------------------------- *)


val star = STAR
fun karrow k1 k2 = KARROW(k1,k2)

fun kprt_exactly k = (write "<kind>")

(* ---------------------------  Types  -------------------------- *)

fun tvar i = TVAR(i)
fun abst n k1 t2 = ABST(n,k1,t2)
fun appt t1 t2 = APPT(t1,t2)
fun arrow t1 t2 = ARROW(t1,t2)
fun all n t1 t2 = ALL(n,t1,t2)
fun some n t1 t2 = SOM(n,t1,t2)
fun top k = TOP(k)
fun rcd l = RCD(l)

fun relocate_above cut off t = let
      fun ra c (TVAR(i))       = tvar (if i>=c then i+off else i)
        | ra c (ABST(n,k1,t2)) = abst n k1 (ra (c+1) t2)
        | ra c (APPT(t1,t2))   = appt (ra c t1) (ra c t2)
        | ra c (ARROW(t1,t2))  = arrow (ra c t1) (ra c t2)
        | ra c (ALL(n,t1,t2))  = all n (ra c t1) (ra (c+1) t2)
        | ra c (SOM(n,t1,t2)) = some n (ra c t1) (ra (c+1) t2)
        | ra c (TOP(k))        = top(k)
        | ra c (RCD(l))        = rcd (map (fn (i,t) => (i,ra c t)) l)
      in ra cut t
      end

fun relocate off t = relocate_above 0 off t

fun tsubst targ i body = let
  fun ts i (TVAR(i'))      = if i=i' 
                                  then (relocate i targ) 
                                else if i < i'
                                  then tvar(i'-1)
                                else tvar(i')
    | ts i (ABST(n,k1,t2)) = abst n k1 (ts (i+1) t2)
    | ts i (APPT(t1,t2))   = appt (ts i t1) (ts i t2)
    | ts i (ARROW(t1,t2))  = arrow (ts i t1) (ts i t2)
    | ts i (ALL(n,t1,t2))  = all n (ts i t1) (ts (i+1) t2)
    | ts i (SOM(n,t1,t2)) = some n (ts i t1) (ts (i+1) t2)
    | ts i (TOP(k))        = top(k)
    | ts i (RCD(l))        = rcd (map (fn (n,t) => (n,ts i t)) l)
  in ts i body
  end

fun tsubst_top targ tbody = tsubst targ 0 tbody

fun occurs_free_in v t = let
  fun s i (TVAR(i'))      = (i = i')
    | s i (ABST(n,k1,t2)) = (s (i+1) t2)
    | s i (APPT(t1,t2))   = (s i t1) orelse (s i t2)
    | s i (ARROW(t1,t2))  = (s i t1) orelse (s i t2)
    | s i (ALL(x,t1,t2))  = (s i t1) orelse (s (i+1) t2)
    | s i (SOM(x,t1,t2)) = (s i t1) orelse (s (i+1) t2)
    | s i (TOP(k))        = false
    | s i (RCD(l))        = forsome (fn (n,t) => (s i t)) l
  in s v t
  end

(* ------------------------------------------------------------------------- *)

fun kprt_exactly (STAR) =
      (write "*")
  | kprt_exactly (KARROW(k1,k2)) =
      (write "(";
       kprt_exactly k1;
       write "->";
       kprt_exactly k2;
       write ")")

and prt_exactly (TVAR(i'))      = 
      (write (makestring i'))
  | prt_exactly (ABST(n,k1,t2)) = 
      (write "((Fun";
       write (Id.tostr n);
       write ":";
       (kprt_exactly k1);
       write ")";
       (prt_exactly t2);
       write ")")
  | prt_exactly (APPT(t1,t2))   = 
      (write "(";
       (prt_exactly t1);
       write " ";
       (prt_exactly t2);
       write ")")
  | prt_exactly (ARROW(t1,t2))  = 
      (write "(";
       (prt_exactly t1);
       write "->";
       (prt_exactly t2);
       write ")")
  | prt_exactly (ALL(n,t1,t2))  = 
      (write "((All ";
       write (Id.tostr n);
       write "<";
       (prt_exactly t1);
       write ")";
       (prt_exactly t2);
       write ")")
  | prt_exactly (SOM(n,t1,t2)) = 
      (write "((Some ";
       write (Id.tostr n);
       write "<";
       (prt_exactly t1);
       write ")";
       (prt_exactly t2);
       write ")")
  | prt_exactly (TOP(k))        = 
      (write "Top(";
       kprt_exactly k;
       write ")")
  | prt_exactly (RCD(l))        = 
      (write "{";
       mapunit_tuple 
          (fn (n,t) => (write (Id.tostr n); write ":"; prt_exactly t))
          (fn()=> write ",") 
          l;
       write "}")

val dummytype = top(star)

end

